# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class TxRequestSerializedType(p.MessageType):
    FIELDS = {
        1: ('signature_index', p.UVarintType, 0),
        2: ('signature', p.BytesType, 0),
        3: ('serialized_tx', p.BytesType, 0),
    }



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .NEMMosaic import NEMMosaic


class NEMTransfer(p.MessageType):
    FIELDS = {
        1: ('recipient', p.UnicodeType, 0),
        2: ('amount', p.UVarintType, 0),
        3: ('payload', p.BytesType, 0),
        4: ('public_key', p.BytesType, 0),
        5: ('mosaics', NEMMosaic, p.FLAG_REPEATED),
    }



# This file is part of the TREZOR project.
#
# Copyright (C) 2016-2017 Pavol Rusnak <stick@satoshilabs.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from .common import *
import trezorlib.ckd_public as bip32
from trezorlib import messages as proto


class TestMsgGetaddressSegwit(TrezorTest):

    def test_show_segwit(self):
        self.setup_mnemonic_allallall()
        assert self.client.get_address("Testnet", self.client.expand_path("49'/1'/0'/1/0"), True, None, script_type=proto.InputScriptType.SPENDP2SHWITNESS) == '2N1LGaGg836mqSQqiuUBLfcyGBhyZbremDX'
        assert self.client.get_address("Testnet", self.client.expand_path("49'/1'/0'/0/0"), False, None, script_type=proto.InputScriptType.SPENDP2SHWITNESS) == '2N4Q5FhU2497BryFfUgbqkAJE87aKHUhXMp'
        assert self.client.get_address("Testnet", self.client.expand_path("44'/1'/0'/0/0"), False, None, script_type=proto.InputScriptType.SPENDP2SHWITNESS) == '2N6UeBoqYEEnybg4cReFYDammpsyDw8R2Mc'
        assert self.client.get_address("Testnet", self.client.expand_path("44'/1'/0'/0/0"), False, None, script_type=proto.InputScriptType.SPENDADDRESS) == 'mvbu1Gdy8SUjTenqerxUaZyYjmveZvt33q'

    def test_show_multisig_3(self):
        self.setup_mnemonic_allallall()
        nodes = map(lambda index: self.client.get_public_node(self.client.expand_path("999'/1'/%d'" % index)), range(1, 4))
        multisig1 = proto.MultisigRedeemScriptType(
            pubkeys=list(map(lambda n: proto.HDNodePathType(node=bip32.deserialize(n.xpub), address_n=[2, 0]), nodes)),
            signatures=[b'', b'', b''],
            m=2,
        )
        # multisig2 = proto.MultisigRedeemScriptType(
        #     pubkeys=map(lambda n: proto.HDNodePathType(node=bip32.deserialize(n.xpub), address_n=[2, 1]), nodes),
        #     signatures=[b'', b'', b''],
        #     m=2,
        # )
        for i in [1, 2, 3]:
            assert self.client.get_address("Testnet", self.client.expand_path("999'/1'/%d'/2/0" % i), False, multisig1, script_type=proto.InputScriptType.SPENDP2SHWITNESS) == '2N2MxyAfifVhb3AMagisxaj3uij8bfXqf4Y'



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class NEMProvisionNamespace(p.MessageType):
    FIELDS = {
        1: ('namespace', p.UnicodeType, 0),
        2: ('parent', p.UnicodeType, 0),
        3: ('sink', p.UnicodeType, 0),
        4: ('fee', p.UVarintType, 0),
    }



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from .common import *
from trezorlib import messages as proto
from mnemonic import Mnemonic


@pytest.mark.skip_t2
class TestMsgResetDeviceSkipbackup(TrezorTest):

    def test_reset_device_skip_backup(self):

        external_entropy = b'zlutoucky kun upel divoke ody' * 2
        strength = 128

        ret = self.client.call_raw(proto.ResetDevice(
            display_random=False,
            strength=strength,
            passphrase_protection=False,
            pin_protection=False,
            language='english',
            label='test',
            skip_backup=True
        ))

        # Provide entropy
        assert isinstance(ret, proto.EntropyRequest)
        internal_entropy = self.client.debug.read_reset_entropy()
        ret = self.client.call_raw(proto.EntropyAck(entropy=external_entropy))
        assert isinstance(ret, proto.Success)

        # Check if device is properly initialized
        resp = self.client.call_raw(proto.Initialize())
        assert resp.initialized is True
        assert resp.needs_backup is True

        # Generate mnemonic locally
        entropy = generate_entropy(strength, internal_entropy, external_entropy)
        expected_mnemonic = Mnemonic('english').to_mnemonic(entropy)

        # start Backup workflow
        ret = self.client.call_raw(proto.BackupDevice())

        mnemonic = []
        for _ in range(strength // 32 * 3):
            assert isinstance(ret, proto.ButtonRequest)
            mnemonic.append(self.client.debug.read_reset_word())
            self.client.debug.press_yes()
            self.client.call_raw(proto.ButtonAck())

        mnemonic = ' '.join(mnemonic)

        # Compare that device generated proper mnemonic for given entropies
        assert mnemonic == expected_mnemonic

        mnemonic = []
        for _ in range(strength // 32 * 3):
            assert isinstance(ret, proto.ButtonRequest)
            mnemonic.append(self.client.debug.read_reset_word())
            self.client.debug.press_yes()
            resp = self.client.call_raw(proto.ButtonAck())

        assert isinstance(resp, proto.Success)

        mnemonic = ' '.join(mnemonic)

        # Compare that second pass printed out the same mnemonic once again
        assert mnemonic == expected_mnemonic

        # start backup again - should fail
        ret = self.client.call_raw(proto.BackupDevice())
        assert isinstance(ret, proto.Failure)

    def test_reset_device_skip_backup_break(self):

        external_entropy = b'zlutoucky kun upel divoke ody' * 2
        strength = 128

        ret = self.client.call_raw(proto.ResetDevice(
            display_random=False,
            strength=strength,
            passphrase_protection=False,
            pin_protection=False,
            language='english',
            label='test',
            skip_backup=True
        ))

        # Provide entropy
        assert isinstance(ret, proto.EntropyRequest)
        ret = self.client.call_raw(proto.EntropyAck(entropy=external_entropy))
        assert isinstance(ret, proto.Success)

        # Check if device is properly initialized
        resp = self.client.call_raw(proto.Initialize())
        assert resp.initialized is True
        assert resp.needs_backup is True

        # start Backup workflow
        ret = self.client.call_raw(proto.BackupDevice())

        # send Initialize -> break workflow
        ret = self.client.call_raw(proto.Initialize())
        assert isinstance(resp, proto.Features)

        # start backup again - should fail
        ret = self.client.call_raw(proto.BackupDevice())
        assert isinstance(ret, proto.Failure)

    def test_initialized_device_backup_fail(self):
        self.setup_mnemonic_nopin_nopassphrase()
        ret = self.client.call_raw(proto.BackupDevice())
        assert isinstance(ret, proto.Failure)



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class EncryptedMessage(p.MessageType):
    FIELDS = {
        1: ('nonce', p.BytesType, 0),
        2: ('message', p.BytesType, 0),
        3: ('hmac', p.BytesType, 0),
    }
    MESSAGE_WIRE_TYPE = 50



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .NEMMosaicDefinition import NEMMosaicDefinition


class NEMMosaicCreation(p.MessageType):
    FIELDS = {
        1: ('definition', NEMMosaicDefinition, 0),
        2: ('sink', p.UnicodeType, 0),
        3: ('fee', p.UVarintType, 0),
    }



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from .common import *
import trezorlib.ckd_public as bip32


class TestMsgGetpublickeyCurve(TrezorTest):

    def test_default_curve(self):
        self.setup_mnemonic_nopin_nopassphrase()
        assert hexlify(self.client.get_public_node([0x80000000 | 111, 42]).node.public_key).decode() == '02e7fcec053f0df94d88c86447970743e8a1979d242d09338dcf8687a9966f7fbc'
        assert hexlify(self.client.get_public_node([0x80000000 | 111, 0x80000000 | 42]).node.public_key).decode() == '03ce7b690969d773ba9ed212464eb2b534b87b9b8a9383300bddabe1f093f79220'

    def test_secp256k1_curve(self):
        self.setup_mnemonic_nopin_nopassphrase()
        assert hexlify(self.client.get_public_node([0x80000000 | 111, 42], ecdsa_curve_name='secp256k1').node.public_key).decode() == '02e7fcec053f0df94d88c86447970743e8a1979d242d09338dcf8687a9966f7fbc'
        assert hexlify(self.client.get_public_node([0x80000000 | 111, 0x80000000 | 42], ecdsa_curve_name='secp256k1').node.public_key).decode() == '03ce7b690969d773ba9ed212464eb2b534b87b9b8a9383300bddabe1f093f79220'

    def test_nist256p1_curve(self):
        self.setup_mnemonic_nopin_nopassphrase()
        assert hexlify(self.client.get_public_node([0x80000000 | 111, 42], ecdsa_curve_name='nist256p1').node.public_key).decode() == '02a9ce59b32bd64a70bc52aca96e5d09af65c6b9593ba2a60af8fccfe1437f2129'
        assert hexlify(self.client.get_public_node([0x80000000 | 111, 0x80000000 | 42], ecdsa_curve_name='nist256p1').node.public_key).decode() == '026fe35d8afed67dbf0561a1d32922e8ad0cd0d86effbc82be970cbed7d9bab2c2'

    def test_ed25519_curve(self):
        self.setup_mnemonic_nopin_nopassphrase()
        assert hexlify(self.client.get_public_node([0x80000000 | 111, 42], ecdsa_curve_name='ed25519').node.public_key).decode() == '001d9a1e56f69828d44ec96dad345678411976d3ea6d290fe3ae8032c47699ce15'
        assert hexlify(self.client.get_public_node([0x80000000 | 111, 0x80000000 | 42], ecdsa_curve_name='ed25519').node.public_key).decode() == '0069a14b478e508eab6e93303f4e6f5c50b8136627830f2ed5c3a835fc6c0ea2b7'



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class CosiCommitment(p.MessageType):
    FIELDS = {
        1: ('commitment', p.BytesType, 0),
        2: ('pubkey', p.BytesType, 0),
    }
    MESSAGE_WIRE_TYPE = 72



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class RecoveryDevice(p.MessageType):
    FIELDS = {
        1: ('word_count', p.UVarintType, 0),
        2: ('passphrase_protection', p.BoolType, 0),
        3: ('pin_protection', p.BoolType, 0),
        4: ('language', p.UnicodeType, 0),  # default='english'
        5: ('label', p.UnicodeType, 0),
        6: ('enforce_wordlist', p.BoolType, 0),
        8: ('type', p.UVarintType, 0),
        9: ('u2f_counter', p.UVarintType, 0),
        10: ('dry_run', p.BoolType, 0),
    }
    MESSAGE_WIRE_TYPE = 45



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .HDNodeType import HDNodeType


class PublicKey(p.MessageType):
    FIELDS = {
        1: ('node', HDNodeType, 0),  # required
        2: ('xpub', p.UnicodeType, 0),
    }
    MESSAGE_WIRE_TYPE = 12



# Automatically generated by pb2py
from __future__ import absolute_import

from .CoinType import *
from .HDNodePathType import *
from .HDNodeType import *
from .IdentityType import *
from .MultisigRedeemScriptType import *
from .NEMAggregateModification import *
from .NEMCosignatoryModification import *
from .NEMImportanceTransfer import *
from .NEMMosaic import *
from .NEMMosaicCreation import *
from .NEMMosaicDefinition import *
from .NEMMosaicSupplyChange import *
from .NEMProvisionNamespace import *
from .NEMTransactionCommon import *
from .NEMTransfer import *
from .TransactionType import *
from .TxInputType import *
from .TxOutputBinType import *
from .TxOutputType import *
from .TxRequestDetailsType import *
from .TxRequestSerializedType import *
from . import FailureType
from . import OutputScriptType
from . import InputScriptType
from . import RequestType
from . import ButtonRequestType
from . import PinMatrixRequestType
from . import RecoveryDeviceType
from . import WordRequestType
from . import NEMMosaicLevy
from . import NEMSupplyChangeType
from . import NEMModificationType
from . import NEMImportanceTransferMode
from .Address import *
from .ApplyFlags import *
from .ApplySettings import *
from .BackupDevice import *
from .ButtonAck import *
from .ButtonRequest import *
from .Cancel import *
from .ChangePin import *
from .CipherKeyValue import *
from .CipheredKeyValue import *
from .ClearSession import *
from .CosiCommit import *
from .CosiCommitment import *
from .CosiSign import *
from .CosiSignature import *
from .DebugLinkDecision import *
from .DebugLinkFlashErase import *
from .DebugLinkGetState import *
from .DebugLinkLog import *
from .DebugLinkMemory import *
from .DebugLinkMemoryRead import *
from .DebugLinkMemoryWrite import *
from .DebugLinkState import *
from .DebugLinkStop import *
from .DecryptMessage import *
from .DecryptedMessage import *
from .ECDHSessionKey import *
from .EncryptMessage import *
from .EncryptedMessage import *
from .Entropy import *
from .EntropyAck import *
from .EntropyRequest import *
from .EstimateTxSize import *
from .EthereumAddress import *
from .EthereumGetAddress import *
from .EthereumMessageSignature import *
from .EthereumSignMessage import *
from .EthereumSignTx import *
from .EthereumTxAck import *
from .EthereumTxRequest import *
from .EthereumVerifyMessage import *
from .Failure import *
from .Features import *
from .FirmwareErase import *
from .FirmwareRequest import *
from .FirmwareUpload import *
from .GetAddress import *
from .GetECDHSessionKey import *
from .GetEntropy import *
from .GetFeatures import *
from .GetPublicKey import *
from .Initialize import *
from .LoadDevice import *
from .MessageSignature import *
from .NEMAddress import *
from .NEMDecryptMessage import *
from .NEMDecryptedMessage import *
from .NEMGetAddress import *
from .NEMSignTx import *
from .NEMSignedTx import *
from .PassphraseAck import *
from .PassphraseRequest import *
from .PassphraseStateAck import *
from .PassphraseStateRequest import *
from .PinMatrixAck import *
from .PinMatrixRequest import *
from .Ping import *
from .PublicKey import *
from .RecoveryDevice import *
from .ResetDevice import *
from .SelfTest import *
from .SetU2FCounter import *
from .SignIdentity import *
from .SignMessage import *
from .SignTx import *
from .SignedIdentity import *
from .SimpleSignTx import *
from .Success import *
from .TxAck import *
from .TxRequest import *
from .TxSize import *
from .VerifyMessage import *
from .WipeDevice import *
from .WordAck import *
from .WordRequest import *
from . import MessageType



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .NEMAggregateModification import NEMAggregateModification
from .NEMImportanceTransfer import NEMImportanceTransfer
from .NEMMosaicCreation import NEMMosaicCreation
from .NEMMosaicSupplyChange import NEMMosaicSupplyChange
from .NEMProvisionNamespace import NEMProvisionNamespace
from .NEMTransactionCommon import NEMTransactionCommon
from .NEMTransfer import NEMTransfer


class NEMSignTx(p.MessageType):
    FIELDS = {
        1: ('transaction', NEMTransactionCommon, 0),
        2: ('multisig', NEMTransactionCommon, 0),
        3: ('transfer', NEMTransfer, 0),
        4: ('cosigning', p.BoolType, 0),
        5: ('provision_namespace', NEMProvisionNamespace, 0),
        6: ('mosaic_creation', NEMMosaicCreation, 0),
        7: ('supply_change', NEMMosaicSupplyChange, 0),
        8: ('aggregate_modification', NEMAggregateModification, 0),
        9: ('importance_transfer', NEMImportanceTransfer, 0),
    }
    MESSAGE_WIRE_TYPE = 69



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from .common import *
from trezorlib import messages as proto
import trezorlib.ckd_public as bip32


class TestMsgGetaddress(TrezorTest):

    def test_btc(self):
        self.setup_mnemonic_nopin_nopassphrase()
        assert self.client.get_address('Bitcoin', []) == '1EfKbQupktEMXf4gujJ9kCFo83k1iMqwqK'
        assert self.client.get_address('Bitcoin', [1]) == '1CK7SJdcb8z9HuvVft3D91HLpLC6KSsGb'
        assert self.client.get_address('Bitcoin', [0, -1]) == '1JVq66pzRBvqaBRFeU9SPVvg3er4ZDgoMs'
        assert self.client.get_address('Bitcoin', [-9, 0]) == '1F4YdQdL9ZQwvcNTuy5mjyQxXkyCfMcP2P'
        assert self.client.get_address('Bitcoin', [0, 9999999]) == '1GS8X3yc7ntzwGw9vXwj9wqmBWZkTFewBV'

    def test_ltc(self):
        self.setup_mnemonic_nopin_nopassphrase()
        assert self.client.get_address('Litecoin', []) == 'LYtGrdDeqYUQnTkr5sHT2DKZLG7Hqg7HTK'
        assert self.client.get_address('Litecoin', [1]) == 'LKRGNecThFP3Q6c5fosLVA53Z2hUDb1qnE'
        assert self.client.get_address('Litecoin', [0, -1]) == 'LcinMK8pVrAtpz7Qpc8jfWzSFsDLgLYfG6'
        assert self.client.get_address('Litecoin', [-9, 0]) == 'LZHVtcwAEDf1BR4d67551zUijyLUpDF9EX'
        assert self.client.get_address('Litecoin', [0, 9999999]) == 'Laf5nGHSCT94C5dK6fw2RxuXPiw2ZuRR9S'

    def test_tbtc(self):
        self.setup_mnemonic_nopin_nopassphrase()
        assert self.client.get_address('Testnet', [111, 42]) == 'moN6aN6NP1KWgnPSqzrrRPvx2x1UtZJssa'

    def test_bch(self):
        self.setup_mnemonic_allallall()
        assert self.client.get_address('Bcash', self.client.expand_path("44'/145'/0'/0/0")) == '1MH9KKcvdCTY44xVDC2k3fjBbX5Cz29N1q'
        assert self.client.get_address('Bcash', self.client.expand_path("44'/145'/0'/0/1")) == '1LRspCZNFJcbuNKQkXgHMDucctFRQya5a3'
        assert self.client.get_address('Bcash', self.client.expand_path("44'/145'/0'/1/0")) == '1HADRPJpgqBzThepERpVXNi6qRgiLQRNoE'

    def test_bch_multisig(self):
        self.setup_mnemonic_allallall()
        xpubs = []
        for n in map(lambda index: self.client.get_public_node(self.client.expand_path("44'/145'/" + str(index) + "'")), range(1, 4)):
            xpubs.append(n.xpub)

        def getmultisig(chain, nr, signatures=[b'', b'', b''], xpubs=xpubs):
            return proto.MultisigRedeemScriptType(
                pubkeys=list(map(lambda xpub: proto.HDNodePathType(node=bip32.deserialize(xpub), address_n=[chain, nr]), xpubs)),
                signatures=signatures,
                m=2,
            )
        for nr in range(1, 4):
            assert self.client.get_address('Bcash', self.client.expand_path("44'/145'/" + str(nr) + "'/0/0"), show_display=(nr == 1), multisig=getmultisig(0, 0)) == '33Ju286QvonBz5N1V754ZekQv4GLJqcc5R'
            assert self.client.get_address('Bcash', self.client.expand_path("44'/145'/" + str(nr) + "'/1/0"), show_display=(nr == 1), multisig=getmultisig(1, 0)) == '3CPtPpL5mGAPdxUeUDfm2RNdWoSN9dKpXE'

    def test_public_ckd(self):
        self.setup_mnemonic_nopin_nopassphrase()

        node = self.client.get_public_node([]).node
        node_sub1 = self.client.get_public_node([1]).node
        node_sub2 = bip32.public_ckd(node, [1])

        assert node_sub1.chain_code == node_sub2.chain_code
        assert node_sub1.public_key == node_sub2.public_key

        address1 = self.client.get_address('Bitcoin', [1])
        address2 = bip32.get_address(node_sub2, 0)

        assert address2 == '1CK7SJdcb8z9HuvVft3D91HLpLC6KSsGb'
        assert address1 == address2



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class ApplyFlags(p.MessageType):
    FIELDS = {
        1: ('flags', p.UVarintType, 0),
    }
    MESSAGE_WIRE_TYPE = 28



from .tx_api import TxApiBitcoin, TxApiTestnet, TxApiLitecoin, TxApiZcash, TxApiDash, TxApiBcash, TxApiDecredTestnet, TxApiDogecoin, TxApiMonacoin

coins_slip44 = {
    'Bitcoin': 0,
    'Testnet': 1,
    'Decred Testnet': 1,
    'Litecoin': 2,
    'Dogecoin': 3,
    'Dash': 5,
    'Namecoin': 7,
    'Monacoin': 22,
    'Decred': 42,
    'Ether': 60,
    'EtherClassic': 61,
    'Zcash': 133,
    'Bcash': 145,
}

coins_txapi = {
    'Bitcoin': TxApiBitcoin,
    'Testnet': TxApiTestnet,
    'Litecoin': TxApiLitecoin,
    'Dash': TxApiDash,
    'Zcash': TxApiZcash,
    'Bcash': TxApiBcash,
    'Decred Testnet': TxApiDecredTestnet,
    'Dogecoin': TxApiDogecoin,
    'Monacoin': TxApiMonacoin,
}



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .CoinType import CoinType


class Features(p.MessageType):
    FIELDS = {
        1: ('vendor', p.UnicodeType, 0),
        2: ('major_version', p.UVarintType, 0),
        3: ('minor_version', p.UVarintType, 0),
        4: ('patch_version', p.UVarintType, 0),
        5: ('bootloader_mode', p.BoolType, 0),
        6: ('device_id', p.UnicodeType, 0),
        7: ('pin_protection', p.BoolType, 0),
        8: ('passphrase_protection', p.BoolType, 0),
        9: ('language', p.UnicodeType, 0),
        10: ('label', p.UnicodeType, 0),
        11: ('coins', CoinType, p.FLAG_REPEATED),
        12: ('initialized', p.BoolType, 0),
        13: ('revision', p.BytesType, 0),
        14: ('bootloader_hash', p.BytesType, 0),
        15: ('imported', p.BoolType, 0),
        16: ('pin_cached', p.BoolType, 0),
        17: ('passphrase_cached', p.BoolType, 0),
        18: ('firmware_present', p.BoolType, 0),
        19: ('needs_backup', p.BoolType, 0),
        20: ('flags', p.UVarintType, 0),
        21: ('model', p.UnicodeType, 0),
        22: ('fw_major', p.UVarintType, 0),
        23: ('fw_minor', p.UVarintType, 0),
        24: ('fw_patch', p.UVarintType, 0),
        25: ('fw_vendor', p.UnicodeType, 0),
        26: ('fw_vendor_keys', p.BytesType, 0),
        100: ('pin_remaining_zones', p.UVarintType, 0),
    }
    MESSAGE_WIRE_TYPE = 17



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from .common import *
from trezorlib import messages as proto


class TestMsgEthereumSigntx(TrezorTest):

    def test_ethereum_signtx_erc20_token(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.EthereumTxRequest(data_length=None),
            ])

            data = bytearray()
            # method id signalizing `transfer(address _to, uint256 _value)` function
            data.extend(unhexlify('a9059cbb'))
            # 1st function argument (to - the receiver)
            data.extend(unhexlify('000000000000000000000000574bbb36871ba6b78e27f4b4dcfb76ea0091880b'))
            # 2nd function argument (value - amount to be transferred)
            data.extend(unhexlify('0000000000000000000000000000000000000000000000000000000000000123'))

            sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=0,
                gas_price=20,
                gas_limit=20,
                # ALTS token address
                to=b'\x63\x8a\xc1\x49\xea\x8e\xf9\xa1\x28\x6c\x41\xb9\x77\x01\x7a\xa7\x35\x9e\x6c\xfa',
                chain_id=1,
                # value needs to be 0, token value is set in the contract (data)
                value=0,
                data=data)

            # taken from T1 might not be 100% correct but still better than nothing
            assert hexlify(sig_r) == b'28bf1b621be9a85d2905fa36511dfbd52ec4b67ba4ad6cb2bd08753c72b93b77'
            assert hexlify(sig_s) == b'2fa605244f80a56cb438df55eb9835489288ec2c0ac0280ada2ccaccfe2b7e38'

    def test_ethereum_signtx_nodata(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.EthereumTxRequest(data_length=None),  # v,r,s checked with assert
            ])

            sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=0,
                gas_price=20,
                gas_limit=20,
                to=unhexlify('1d1c328764a41bda0492b66baa30c4a339ff85ef'),
                value=10)

        assert sig_v == 27
        assert hexlify(sig_r) == b'9b61192a161d056c66cfbbd331edb2d783a0193bd4f65f49ee965f791d898f72'
        assert hexlify(sig_s) == b'49c0bbe35131592c6ed5c871ac457feeb16a1493f64237387fab9b83c1a202f7'

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.EthereumTxRequest(data_length=None),
            ])

            sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=123456,
                gas_price=20000,
                gas_limit=20000,
                to=unhexlify('1d1c328764a41bda0492b66baa30c4a339ff85ef'),
                value=12345678901234567890)
        assert sig_v == 28
        assert hexlify(sig_r) == b'6de597b8ec1b46501e5b159676e132c1aa78a95bd5892ef23560a9867528975a'
        assert hexlify(sig_s) == b'6e33c4230b1ecf96a8dbb514b4aec0a6d6ba53f8991c8143f77812aa6daa993f'

    def test_ethereum_signtx_data(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.EthereumTxRequest(data_length=None),
            ])

            sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=0,
                gas_price=20,
                gas_limit=20,
                to=unhexlify('1d1c328764a41bda0492b66baa30c4a339ff85ef'),
                value=10,
                data=b'abcdefghijklmnop' * 16)
        assert sig_v == 28
        assert hexlify(sig_r) == b'6da89ed8627a491bedc9e0382f37707ac4e5102e25e7a1234cb697cedb7cd2c0'
        assert hexlify(sig_s) == b'691f73b145647623e2d115b208a7c3455a6a8a83e3b4db5b9c6d9bc75825038a'

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.EthereumTxRequest(data_length=1024, signature_r=None, signature_s=None, signature_v=None),
                proto.EthereumTxRequest(data_length=1024),
                proto.EthereumTxRequest(data_length=1024),
                proto.EthereumTxRequest(data_length=3),
                proto.EthereumTxRequest(),
            ])

            sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=123456,
                gas_price=20000,
                gas_limit=20000,
                to=unhexlify('1d1c328764a41bda0492b66baa30c4a339ff85ef'),
                value=12345678901234567890,
                data=b'ABCDEFGHIJKLMNOP' * 256 + b'!!!')
        assert sig_v == 28
        assert hexlify(sig_r) == b'4e90b13c45c6a9bf4aaad0e5427c3e62d76692b36eb727c78d332441b7400404'
        assert hexlify(sig_s) == b'3ff236e7d05f0f9b1ee3d70599bb4200638f28388a8faf6bb36db9e04dc544be'

    def test_ethereum_signtx_message(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.EthereumTxRequest(data_length=1024, signature_r=None, signature_s=None, signature_v=None),
                proto.EthereumTxRequest(data_length=1024),
                proto.EthereumTxRequest(data_length=1024),
                proto.EthereumTxRequest(data_length=3),
                proto.EthereumTxRequest(),
            ])

            sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=0,
                gas_price=20000,
                gas_limit=20000,
                to=unhexlify('1d1c328764a41bda0492b66baa30c4a339ff85ef'),
                value=0,
                data=b'ABCDEFGHIJKLMNOP' * 256 + b'!!!')
        assert sig_v == 28
        assert hexlify(sig_r) == b'070e9dafda4d9e733fa7b6747a75f8a4916459560efb85e3e73cd39f31aa160d'
        assert hexlify(sig_s) == b'7842db33ef15c27049ed52741db41fe3238a6fa3a6a0888fcfb74d6917600e41'

    def test_ethereum_signtx_newcontract(self):
        self.setup_mnemonic_nopin_nopassphrase()

        # contract creation without data should fail.
        with pytest.raises(Exception):
            self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=123456,
                gas_price=20000,
                gas_limit=20000,
                to='',
                value=12345678901234567890
            )

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.EthereumTxRequest(data_length=1024, signature_r=None, signature_s=None, signature_v=None),
                proto.EthereumTxRequest(data_length=1024),
                proto.EthereumTxRequest(data_length=1024),
                proto.EthereumTxRequest(data_length=3),
                proto.EthereumTxRequest(),
            ])

            sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=0,
                gas_price=20000,
                gas_limit=20000,
                to='',
                value=12345678901234567890,
                data=b'ABCDEFGHIJKLMNOP' * 256 + b'!!!')
        assert sig_v == 28
        assert hexlify(sig_r) == b'b401884c10ae435a2e792303b5fc257a09f94403b2883ad8c0ac7a7282f5f1f9'
        assert hexlify(sig_s) == b'4742fc9e6a5fa8db3db15c2d856914a7f3daab21603a6c1ce9e9927482f8352e'

    def test_ethereum_sanity_checks(self):
        # gas overflow
        with pytest.raises(Exception):
            self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=123456,
                gas_price=0xffffffffffffffffffffffffffffffff,
                gas_limit=0xffffffffffffffffffffffffffffff,
                to=unhexlify('1d1c328764a41bda0492b66baa30c4a339ff85ef'),
                value=12345678901234567890
            )

        # no gas price
        with pytest.raises(Exception):
            self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=123456,
                gas_limit=10000,
                to=unhexlify('1d1c328764a41bda0492b66baa30c4a339ff85ef'),
                value=12345678901234567890
            )

        # no gas limit
        with pytest.raises(Exception):
            self.client.ethereum_sign_tx(
                n=[0, 0],
                nonce=123456,
                gas_price=10000,
                to=unhexlify('1d1c328764a41bda0492b66baa30c4a339ff85ef'),
                value=12345678901234567890
            )

        # no nonce
        with pytest.raises(Exception):
            self.client.ethereum_sign_tx(
                n=[0, 0],
                gas_price=10000,
                gas_limit=123456,
                to=unhexlify('1d1c328764a41bda0492b66baa30c4a339ff85ef'),
                value=12345678901234567890
            )

    def test_ethereum_signtx_nodata_eip155(self):
        self.setup_mnemonic_allallall()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.EthereumTxRequest(),
            ])

            sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
                n=[0x80000000 | 44, 0x80000000 | 1, 0x80000000, 0, 0],
                nonce=0,
                gas_price=20000000000,
                gas_limit=21000,
                to=unhexlify('8ea7a3fccc211ed48b763b4164884ddbcf3b0a98'),
                value=100000000000000000,
                chain_id=3)
        assert sig_v == 41
        assert hexlify(sig_r) == b'a90d0bc4f8d63be69453dd62f2bb5fff53c610000abf956672564d8a654d401a'
        assert hexlify(sig_s) == b'544a2e57bc8b4da18660a1e6036967ea581cc635f5137e3ba97a750867c27cf2'

        sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
            n=[0x80000000 | 44, 0x80000000 | 1, 0x80000000, 0, 0],
            nonce=1,
            gas_price=20000000000,
            gas_limit=21000,
            to=unhexlify('8ea7a3fccc211ed48b763b4164884ddbcf3b0a98'),
            value=100000000000000000,
            chain_id=3)
        assert sig_v == 42
        assert hexlify(sig_r) == b'699428a6950e23c6843f1bf3754f847e64e047e829978df80d55187d19a401ce'
        assert hexlify(sig_s) == b'087343d0a3a2f10842218ffccb146b59a8431b6245ab389fde22dc833f171e6e'

    def test_ethereum_signtx_data_eip155(self):
        self.setup_mnemonic_allallall()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.EthereumTxRequest(),
            ])

            sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
                n=[0x80000000 | 44, 0x80000000 | 1, 0x80000000, 0, 0],
                nonce=2,
                gas_price=20000000000,
                gas_limit=21004,
                to=unhexlify('8ea7a3fccc211ed48b763b4164884ddbcf3b0a98'),
                value=100000000000000000,
                data=b'\0',
                chain_id=3)
        assert sig_v == 42
        assert hexlify(sig_r) == b'ba85b622a8bb82606ba96c132e81fa8058172192d15bc41d7e57c031bca17df4'
        assert hexlify(sig_s) == b'6473b75997634b6f692f8d672193591d299d5bf1c2d6e51f1a14ed0530b91c7d'

        sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
            n=[0x80000000 | 44, 0x80000000 | 1, 0x80000000, 0, 0],
            nonce=3,
            gas_price=20000000000,
            gas_limit=299732,
            to=unhexlify('8ea7a3fccc211ed48b763b4164884ddbcf3b0a98'),
            value=100000000000000000,
            data=b'ABCDEFGHIJKLMNOP' * 256 + b'!!!',
            chain_id=3)
        assert sig_v == 42
        assert hexlify(sig_r) == b'd021c98f92859c8db5e4de2f0e410a8deb0c977eb1a631e323ebf7484bd0d79a'
        assert hexlify(sig_s) == b'2c0e9defc9b1e895dc9520ff25ba3c635b14ad70aa86a5ad6c0a3acb82b569b6'

        sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
            n=[0x80000000 | 44, 0x80000000 | 1, 0x80000000, 0, 0],
            nonce=4,
            gas_price=20000000000,
            gas_limit=21004,
            to=unhexlify('8ea7a3fccc211ed48b763b4164884ddbcf3b0a98'),
            value=0,
            data=b'\0',
            chain_id=3)
        assert sig_v == 42
        assert hexlify(sig_r) == b'dd52f026972a83c56b7dea356836fcfc70a68e3b879cdc8ef2bb5fea23e0a7aa'
        assert hexlify(sig_s) == b'079285fe579c9a2da25c811b1c5c0a74cd19b6301ee42cf20ef7b3b1353f7242'

        sig_v, sig_r, sig_s = self.client.ethereum_sign_tx(
            n=[0x80000000 | 44, 0x80000000 | 1, 0x80000000, 0, 0],
            nonce=5,
            gas_price=0,
            gas_limit=21004,
            to=unhexlify('8ea7a3fccc211ed48b763b4164884ddbcf3b0a98'),
            value=0,
            data=b'\0',
            chain_id=3)
        assert sig_v == 42
        assert hexlify(sig_r) == b'f7505f709d5999343aea3c384034c62d0514336ff6c6af65582006f708f81503'
        assert hexlify(sig_s) == b'44e09e29a4b6247000b46ddc94fe391e94deb2b39ad6ac6398e6db5bec095ba9'



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
# Copyright (C) 2018      Archos S.A.
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import time
import hid
import os

from .protocol_v1 import ProtocolV1
from .protocol_v2 import ProtocolV2
from .transport import Transport, TransportException

DEV_TREZOR1 = (0x534c, 0x0001)
DEV_TREZOR2 = (0x1209, 0x53c1)
DEV_TREZOR2_BL = (0x1209, 0x53c0)

DEV_SAFE_T_MINI_BOOTLOADER = (0x0e79, 0x6001)
DEV_SAFE_T_MINI = (0x0e79, 0x6000)

# Whether to support all devices based on the Trezor protocol, or just the Safe-T mini
MULTIPLE_DEVICE_SUPPORT = False


class HidHandle(object):

    def __init__(self, path):
        self.path = path
        self.count = 0
        self.handle = None

    def open(self):
        if self.count == 0:
            self.handle = hid.device()
            self.handle.open_path(self.path)
            self.handle.set_nonblocking(True)
        self.count += 1

    def close(self):
        if self.count == 1:
            self.handle.close()
        if self.count > 0:
            self.count -= 1


class HidTransport(Transport):
    '''
    HidTransport implements transport over USB HID interface.
    '''

    PATH_PREFIX = 'hid'

    def __init__(self, device, protocol=None, hid_handle=None):
        super(HidTransport, self).__init__()

        if hid_handle is None:
            hid_handle = HidHandle(device['path'])

        if protocol is None:
            # force_v1 = os.environ.get('TREZOR_TRANSPORT_V1', '0')
            force_v1 = True

            if is_protocol_v2_supported(device) and not int(force_v1):
                protocol = ProtocolV2()
            else:
                protocol = ProtocolV1()

        self.device = device
        self.protocol = protocol
        self.hid = hid_handle
        self.hid_version = None

    def __str__(self):
        return self.get_path()

    def get_path(self):
        return "%s:%s" % (self.PATH_PREFIX, self.device['path'].decode())

    @staticmethod
    def enumerate(debug=False):
        devices = []
        for dev in hid.enumerate(0, 0):
            if not is_supported_device(dev):
                continue
            if debug:
                if not is_debuglink(dev):
                    continue
            else:
                if not is_wirelink(dev):
                    continue
            devices.append(HidTransport(dev))
        return devices

    @classmethod
    def find_by_path(cls, path):
        if isinstance(path, str):
            path = path.encode()
        path = path.replace(cls.PATH_PREFIX.encode() + b':', b'')

        for transport in HidTransport.enumerate():
            if path is None or transport.device['path'] == path:
                return transport
        raise TransportException('HID device not found')

    def find_debug(self):
        if isinstance(self.protocol, ProtocolV2):
            # For v2 protocol, lets use the same HID interface, but with a different session
            protocol = ProtocolV2()
            debug = HidTransport(self.device, protocol, self.hid)
            return debug
        if isinstance(self.protocol, ProtocolV1):
            # For v1 protocol, find debug USB interface for the same serial number
            for debug in HidTransport.enumerate(debug=True):
                if debug.device['serial_number'] == self.device['serial_number']:
                    return debug
        raise TransportException('Debug HID device not found')

    def open(self):
        self.hid.open()
        if is_trezor1(self.device):
            self.hid_version = self.probe_hid_version()
        else:
            self.hid_version = 2
        self.protocol.session_begin(self)

    def close(self):
        self.protocol.session_end(self)
        self.hid.close()
        self.hid_version = None

    def read(self):
        return self.protocol.read(self)

    def write(self, msg):
        return self.protocol.write(self, msg)

    def write_chunk(self, chunk):
        if len(chunk) != 64:
            raise TransportException('Unexpected chunk size: %d' % len(chunk))
        if self.hid_version == 2:
            self.hid.handle.write(b'\0' + bytearray(chunk))
        else:
            self.hid.handle.write(chunk)

    def read_chunk(self):
        while True:
            chunk = self.hid.handle.read(64)
            if chunk:
                break
            else:
                time.sleep(0.001)
        if len(chunk) != 64:
            raise TransportException('Unexpected chunk size: %d' % len(chunk))
        return bytearray(chunk)

    def probe_hid_version(self):
        n = self.hid.handle.write([0, 63] + [0xFF] * 63)
        if n == 65:
            return 2
        n = self.hid.handle.write([63] + [0xFF] * 63)
        if n == 64:
            return 1
        raise TransportException('Unknown HID version')


def is_trezor1(dev):
    return (dev['vendor_id'], dev['product_id']) == DEV_TREZOR1


def is_trezor2(dev):
    return (dev['vendor_id'], dev['product_id']) == DEV_TREZOR2


def is_trezor2_bl(dev):
    return (dev['vendor_id'], dev['product_id']) == DEV_TREZOR2_BL


def is_safet_mini(dev):
    return (dev['vendor_id'], dev['product_id']) == DEV_SAFE_T_MINI


def is_safet_mini_bl(dev):
    return (dev['vendor_id'], dev['product_id']) == DEV_SAFE_T_MINI_BOOTLOADER


def is_protocol_v2_supported(dev):
    if MULTIPLE_DEVICE_SUPPORT:
        return is_trezor2(dev)
    else:
        return False


def is_protocol_v1_supported(dev):
    if MULTIPLE_DEVICE_SUPPORT:
        return is_trezor1(dev) | is_trezor2_bl(dev) | is_safet_mini(dev) | is_safet_mini_bl(dev)
    else:
        return is_safet_mini(dev) | is_safet_mini_bl(dev)
  

def is_supported_device(dev):
    return is_protocol_v1_supported(dev) | is_protocol_v2_supported(dev)


def is_wirelink(dev):
    return (dev['usage_page'] == 0xFF00 or dev['interface_number'] == 0)


def is_debuglink(dev):
    return (dev['usage_page'] == 0xFF01 or dev['interface_number'] == 1)



import sys
from functools import reduce
import binascii

from . import ed25519raw


def combine_keys(pks):
    P = [ed25519raw.decodepoint(pk) for pk in pks]
    combine = reduce(ed25519raw.edwards, P)
    return ed25519raw.encodepoint(combine)


def combine_sig(R, sigs):
    S = [ed25519raw.decodeint(si) for si in sigs]
    s = sum(S) % ed25519raw.l
    sig = R + ed25519raw.encodeint(s)
    return sig


def get_nonce(sk, data, ctr):
    h = ed25519raw.H(sk)
    b = ed25519raw.b
    r = ed25519raw.Hint(bytes([h[i] for i in range(b >> 3, b >> 2)]) + data + binascii.unhexlify('%08x' % ctr))
    R = ed25519raw.scalarmult(ed25519raw.B, r)
    return r, ed25519raw.encodepoint(R)


def self_test(digest):

    def to_hex(by):
        return binascii.hexlify(by).decode()

    N = 3
    keyset = [0, 2]

    digest = binascii.unhexlify(digest)
    print('Digest: %s' % to_hex(digest))
    sks = []
    pks = []
    nonces = []
    commits = []
    sigs = []
    for i in range(0, N):
        print('----- Key %d ------' % (i + 1))
        seckey = (chr(0x41 + i) * 32).encode()
        pubkey = ed25519raw.publickey(seckey)
        print('Secret Key: %s' % to_hex(seckey))
        print('Public Key: %s' % to_hex(pubkey))
        sks.append(seckey)
        pks.append(pubkey)
        ctr = 0
        r, R = get_nonce(seckey, digest, ctr)
        print('Local nonce:  %s' % to_hex(ed25519raw.encodeint(r)))
        print('Local commit: %s' % to_hex(R))
        nonces.append(r)
        commits.append(R)

    global_pk = combine_keys([pks[i] for i in keyset])
    global_R = combine_keys([commits[i] for i in keyset])
    print('-----------------')
    print('Global pubkey: %s' % to_hex(global_pk))
    print('Global commit: %s' % to_hex(global_R))
    print('-----------------')

    for i in range(0, N):
        seckey = sks[i]
        pubkey = pks[i]
        r = nonces[i]
        R = commits[i]
        h = ed25519raw.H(seckey)
        b = ed25519raw.b
        a = 2**(b - 2) + sum(2**i * ed25519raw.bit(h, i) for i in range(3, b - 2))
        S = (r + ed25519raw.Hint(global_R + global_pk + digest) * a) % ed25519raw.l
        print('Local sig %d: %s' % (i + 1, to_hex(ed25519raw.encodeint(S))))
        sigs.append(ed25519raw.encodeint(S))

    print('-----------------')
    sig = combine_sig(global_R, [sigs[i] for i in keyset])
    print('Global sig: %s' % to_hex(sig))
    ed25519raw.checkvalid(sig, digest, global_pk)
    print('Valid Signature!')


if __name__ == '__main__':
    if len(sys.argv) > 1:
        self_test(digest=sys.argv[1])
    else:
        self_test(digest='4a5e1e4baab89f3a32518a88c31bc87f618f76673e2cc77ab2127b7afdeda33b')



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class EthereumTxRequest(p.MessageType):
    FIELDS = {
        1: ('data_length', p.UVarintType, 0),
        2: ('signature_v', p.UVarintType, 0),
        3: ('signature_r', p.BytesType, 0),
        4: ('signature_s', p.BytesType, 0),
    }
    MESSAGE_WIRE_TYPE = 59



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class Entropy(p.MessageType):
    FIELDS = {
        1: ('entropy', p.BytesType, 0),  # required
    }
    MESSAGE_WIRE_TYPE = 10



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class EthereumSignTx(p.MessageType):
    FIELDS = {
        1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
        2: ('nonce', p.BytesType, 0),
        3: ('gas_price', p.BytesType, 0),
        4: ('gas_limit', p.BytesType, 0),
        5: ('to', p.BytesType, 0),
        6: ('value', p.BytesType, 0),
        7: ('data_initial_chunk', p.BytesType, 0),
        8: ('data_length', p.UVarintType, 0),
        9: ('chain_id', p.UVarintType, 0),
    }
    MESSAGE_WIRE_TYPE = 58



# Automatically generated by pb2py
Initialize = 0
Ping = 1
Success = 2
Failure = 3
ChangePin = 4
WipeDevice = 5
FirmwareErase = 6
FirmwareUpload = 7
FirmwareRequest = 8
GetEntropy = 9
Entropy = 10
GetPublicKey = 11
PublicKey = 12
LoadDevice = 13
ResetDevice = 14
SignTx = 15
SimpleSignTx = 16
Features = 17
PinMatrixRequest = 18
PinMatrixAck = 19
Cancel = 20
TxRequest = 21
TxAck = 22
CipherKeyValue = 23
ClearSession = 24
ApplySettings = 25
ButtonRequest = 26
ButtonAck = 27
ApplyFlags = 28
GetAddress = 29
Address = 30
SelfTest = 32
BackupDevice = 34
EntropyRequest = 35
EntropyAck = 36
SignMessage = 38
VerifyMessage = 39
MessageSignature = 40
PassphraseRequest = 41
PassphraseAck = 42
PassphraseStateRequest = 77
PassphraseStateAck = 78
EstimateTxSize = 43
TxSize = 44
RecoveryDevice = 45
WordRequest = 46
WordAck = 47
CipheredKeyValue = 48
EncryptMessage = 49
EncryptedMessage = 50
DecryptMessage = 51
DecryptedMessage = 52
SignIdentity = 53
SignedIdentity = 54
GetFeatures = 55
EthereumGetAddress = 56
EthereumAddress = 57
EthereumSignTx = 58
EthereumTxRequest = 59
EthereumTxAck = 60
GetECDHSessionKey = 61
ECDHSessionKey = 62
SetU2FCounter = 63
EthereumSignMessage = 64
EthereumVerifyMessage = 65
EthereumMessageSignature = 66
NEMGetAddress = 67
NEMAddress = 68
NEMSignTx = 69
NEMSignedTx = 70
CosiCommit = 71
CosiCommitment = 72
CosiSign = 73
CosiSignature = 74
NEMDecryptMessage = 75
NEMDecryptedMessage = 76
DebugLinkDecision = 100
DebugLinkGetState = 101
DebugLinkState = 102
DebugLinkStop = 103
DebugLinkLog = 104
DebugLinkMemoryRead = 110
DebugLinkMemory = 111
DebugLinkMemoryWrite = 112
DebugLinkFlashErase = 113



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
# Copyright (C) 2018      Archos S.A.
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import time
import os
import atexit
import usb1

from .protocol_v1 import ProtocolV1
from .protocol_v2 import ProtocolV2
from .transport import Transport, TransportException

DEV_TREZOR1 = (0x534c, 0x0001)
DEV_TREZOR2 = (0x1209, 0x53c1)
DEV_TREZOR2_BL = (0x1209, 0x53c0)

DEV_SAFE_T_MINI_BOOTLOADER = (0x0e79, 0x6001)
DEV_SAFE_T_MINI = (0x0e79, 0x6000)

INTERFACE = 0
ENDPOINT = 1
DEBUG_INTERFACE = 1
DEBUG_ENDPOINT = 2

# Whether to support all devices based on the Trezor protocol, or just the Safe-T mini
MULTIPLE_DEVICE_SUPPORT = False


class WebUsbHandle(object):

    def __init__(self, device):
        self.device = device
        self.count = 0
        self.handle = None

    def open(self, interface):
        if self.count == 0:
            self.handle = self.device.open()
            if self.handle is None:
                raise Exception('Cannot open device')
            self.handle.claimInterface(interface)
            self.count += 1

    def close(self, interface):
        if self.count == 1:
            self.handle.releaseInterface(interface)
            self.handle.close()
        if self.count > 0:
            self.count -= 1


class WebUsbTransport(Transport):
    '''
    WebUsbTransport implements transport over WebUSB interface.
    '''

    PATH_PREFIX = 'webusb'
    context = None

    def __init__(self, device, protocol=None, handle=None, debug=False):
        super(WebUsbTransport, self).__init__()

        if handle is None:
            handle = WebUsbHandle(device)

        if protocol is None:
            # force_v1 = os.environ.get('TREZOR_TRANSPORT_V1', '0')
            force_v1 = True

            if is_protocol_v2_supported(device) and not int(force_v1):
                protocol = ProtocolV2()
            else:
                protocol = ProtocolV1()

        self.device = device
        self.protocol = protocol
        self.handle = handle
        self.debug = debug

    def __str__(self):
        return self.get_path()

    def get_path(self):
        return "%s:%s" % (self.PATH_PREFIX, dev_to_str(self.device))

    @classmethod
    def enumerate(cls):
        if cls.context is None:
            cls.context = usb1.USBContext()
            cls.context.open()
            atexit.register(cls.context.close)
        devices = []
        for dev in cls.context.getDeviceIterator(skip_on_error=True):
            if not is_supported_device(dev):
                continue
            if not is_vendor_class(dev):
                continue
            devices.append(WebUsbTransport(dev))
        return devices

    @classmethod
    def find_by_path(cls, path):
        path = path.replace('%s:' % cls.PATH_PREFIX, '')  # Remove prefix from __str__()
        for transport in WebUsbTransport.enumerate():
            if path is None or dev_to_str(transport.device) == path:
                return transport
        raise TransportException('WebUSB device not found')

    def find_debug(self):
        if isinstance(self.protocol, ProtocolV2):
            # TODO test this
            # For v2 protocol, lets use the same WebUSB interface, but with a different session
            protocol = ProtocolV2()
            debug = WebUsbTransport(self.device, protocol, self.handle)
            return debug
        if isinstance(self.protocol, ProtocolV1):
            # For v1 protocol, find debug USB interface for the same serial number
            protocol = ProtocolV1()
            debug = WebUsbTransport(self.device, protocol, None, True)
            return debug
        raise TransportException('Debug WebUSB device not found')

    def open(self):
        interface = DEBUG_INTERFACE if self.debug else INTERFACE
        self.handle.open(interface)
        self.protocol.session_begin(self)

    def close(self):
        interface = DEBUG_INTERFACE if self.debug else INTERFACE
        self.protocol.session_end(self)
        self.handle.close(interface)

    def read(self):
        return self.protocol.read(self)

    def write(self, msg):
        return self.protocol.write(self, msg)

    def write_chunk(self, chunk):
        endpoint = DEBUG_ENDPOINT if self.debug else ENDPOINT
        if len(chunk) != 64:
            raise TransportException('Unexpected chunk size: %d' % len(chunk))
        self.handle.handle.interruptWrite(endpoint, chunk)

    def read_chunk(self):
        endpoint = DEBUG_ENDPOINT if self.debug else ENDPOINT
        endpoint = 0x80 | endpoint
        while True:
            chunk = self.handle.handle.interruptRead(endpoint, 64)
            if chunk:
                break
            else:
                time.sleep(0.001)
        if len(chunk) != 64:
            raise TransportException('Unexpected chunk size: %d' % len(chunk))
        return bytearray(chunk)


def is_trezor1(dev):
    return (dev.getVendorID(), dev.getProductID()) == DEV_TREZOR1


def is_trezor2(dev):
    return (dev.getVendorID(), dev.getProductID()) == DEV_TREZOR2


def is_trezor2_bl(dev):
    return (dev.getVendorID(), dev.getProductID()) == DEV_TREZOR2_BL


def is_safet_mini(dev):
    return (dev.getVendorID(), dev.getProductID()) == DEV_SAFE_T_MINI


def is_safet_mini_bl(dev):
    return (dev.getVendorID(), dev.getProductID()) == DEV_SAFE_T_MINI_BOOTLOADER


def is_protocol_v2_supported(dev):
    if MULTIPLE_DEVICE_SUPPORT:
        return is_trezor2(dev)
    else:
        return False


def is_protocol_v1_supported(dev):
    if MULTIPLE_DEVICE_SUPPORT:
        return is_trezor1(dev) | is_trezor2_bl(dev) | is_safet_mini(dev) | is_safet_mini_bl(dev)
    else:
        return is_safet_mini(dev) | is_safet_mini_bl(dev)
  

def is_supported_device(dev):
    return is_protocol_v1_supported(dev) | is_protocol_v2_supported(dev)


def is_vendor_class(dev):
    configurationId = 0
    altSettingId = 0
    return dev[configurationId][INTERFACE][altSettingId].getClass() == usb1.libusb1.LIBUSB_CLASS_VENDOR_SPEC


def dev_to_str(dev):
    return ':'.join(str(x) for x in ['%03i' % (dev.getBusNumber(), )] + dev.getPortNumberList())



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
# Copyright (C) 2016      Jochen Hoenicke <hoenicke@gmail.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from . import messages
from . import protobuf

map_type_to_class = {}
map_class_to_type = {}


def build_map():
    for msg_name in dir(messages.MessageType):
        if msg_name.startswith('__'):
            continue

        try:
            msg_class = getattr(messages, msg_name)
        except AttributeError:
            raise ValueError("Implementation of protobuf message '%s' is missing" % msg_name)

        if msg_class.MESSAGE_WIRE_TYPE != getattr(messages.MessageType, msg_name):
            raise ValueError("Inconsistent wire type and MessageType record for '%s'" % msg_class)

        register_message(msg_class)


def register_message(msg_class):
    if msg_class.MESSAGE_WIRE_TYPE in map_type_to_class:
        raise Exception("Message for wire type %s is already registered by %s" %
                        (msg_class.MESSAGE_WIRE_TYPE, get_class(msg_class.MESSAGE_WIRE_TYPE)))

    map_class_to_type[msg_class] = msg_class.MESSAGE_WIRE_TYPE
    map_type_to_class[msg_class.MESSAGE_WIRE_TYPE] = msg_class


def get_type(msg):
    return map_class_to_type[msg.__class__]


def get_class(t):
    return map_type_to_class[t]


build_map()



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class EstimateTxSize(p.MessageType):
    FIELDS = {
        1: ('outputs_count', p.UVarintType, 0),  # required
        2: ('inputs_count', p.UVarintType, 0),  # required
        3: ('coin_name', p.UnicodeType, 0),  # default='Bitcoin'
    }
    MESSAGE_WIRE_TYPE = 43



# This file is part of the TREZOR project.
#
# Copyright (C) 2016-2017 Pavol Rusnak <stick@satoshilabs.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from .common import *


@pytest.mark.skip_t2
class TestMsgEthereumVerifymessage(TrezorTest):

    def test_verify(self):
        self.setup_mnemonic_nopin_nopassphrase()
        res = self.client.ethereum_verify_message(
            unhexlify('cb3864960e8db1a751212c580af27ee8867d688f'),
            unhexlify('95b64a7b3aa492f0cc1668a24097004562cc2b4f0e755e3c0d60dd791b9f9e285f95b618258ff97036b8419d0a0dd1af3751c625b4d248ee6deff84eba21b8ee1c'),
            'This is an example of a signed message.'
        )
        assert res is True

    def test_verify_long(self):
        self.setup_mnemonic_nopin_nopassphrase()
        ret = self.client.ethereum_verify_message(
            unhexlify('cb3864960e8db1a751212c580af27ee8867d688f'),
            unhexlify('70d03c8447b64489e80ae44ce4f1a543e8eb5dd9e9a19c4743ce95fbd9b8234b2d2a16db87cee857f5b474107ad2c0c0c86118f8a33d5df3d98b766be92d71331b'),
            'VeryLongMessage!' * 64
        )
        assert ret is True



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class ApplySettings(p.MessageType):
    FIELDS = {
        1: ('language', p.UnicodeType, 0),
        2: ('label', p.UnicodeType, 0),
        3: ('use_passphrase', p.BoolType, 0),
        4: ('homescreen', p.BytesType, 0),
    }
    MESSAGE_WIRE_TYPE = 25



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from .common import *
from trezorlib import messages as proto


@pytest.mark.skip_t2
class TestMsgRecoverydeviceDryrun(TrezorTest):

    def recovery_loop(self, mnemonic, result):
        ret = self.client.call_raw(proto.RecoveryDevice(word_count=12,
                                   passphrase_protection=False,
                                   pin_protection=False,
                                   label='label',
                                   language='english',
                                   enforce_wordlist=True,
                                   dry_run=True))

        fakes = 0
        for _ in range(int(12 * 2)):
            assert isinstance(ret, proto.WordRequest)
            (word, pos) = self.client.debug.read_recovery_word()

            if pos != 0:
                ret = self.client.call_raw(proto.WordAck(word=mnemonic[pos - 1]))
                mnemonic[pos - 1] = None
            else:
                ret = self.client.call_raw(proto.WordAck(word=word))
                fakes += 1

            print(mnemonic)

        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()

        ret = self.client.call_raw(proto.ButtonAck())
        assert isinstance(ret, result)

    def test_correct_notsame(self):
        self.setup_mnemonic_nopin_nopassphrase()
        mnemonic = ['all'] * 12
        self.recovery_loop(mnemonic, proto.Failure)

    def test_correct_same(self):
        self.setup_mnemonic_nopin_nopassphrase()
        mnemonic = self.mnemonic12.split(' ')
        self.recovery_loop(mnemonic, proto.Success)

    def test_incorrect(self):
        self.setup_mnemonic_nopin_nopassphrase()
        mnemonic = ['stick'] * 12
        self.recovery_loop(mnemonic, proto.Failure)



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class SignTx(p.MessageType):
    FIELDS = {
        1: ('outputs_count', p.UVarintType, 0),  # required
        2: ('inputs_count', p.UVarintType, 0),  # required
        3: ('coin_name', p.UnicodeType, 0),  # default='Bitcoin'
        4: ('version', p.UVarintType, 0),  # default=1
        5: ('lock_time', p.UVarintType, 0),  # default=0
        6: ('decred_expiry', p.UVarintType, 0),
        7: ('preblock_hash', p.BytesType, 0), #default=0
    }
    MESSAGE_WIRE_TYPE = 15



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .MultisigRedeemScriptType import MultisigRedeemScriptType


class TxInputType(p.MessageType):
    FIELDS = {
        1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
        2: ('prev_hash', p.BytesType, 0),  # required
        3: ('prev_index', p.UVarintType, 0),  # required
        4: ('script_sig', p.BytesType, 0),
        5: ('sequence', p.UVarintType, 0),  # default=4294967295
        6: ('script_type', p.UVarintType, 0),  # default=0
        7: ('multisig', MultisigRedeemScriptType, 0),
        8: ('amount', p.UVarintType, 0),
        9: ('decred_tree', p.UVarintType, 0),
        10: ('decred_script_version', p.UVarintType, 0),
    }



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class TxRequestDetailsType(p.MessageType):
    FIELDS = {
        1: ('request_index', p.UVarintType, 0),
        2: ('tx_hash', p.BytesType, 0),
        3: ('extra_data_len', p.UVarintType, 0),
        4: ('extra_data_offset', p.UVarintType, 0),
    }



# This file is part of the TREZOR project.
#
# Copyright (C) 2017 Saleem Rashid <trezor@saleemrashid.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from .common import *

from trezorlib import messages as proto

# tx hash: 209368053ac61969b6838ceb7e31badeb622ed6aa42d6c58365c42ad1a11e19d
SIGNATURE_TESTNET_SIMPLE = unhexlify(
    "9cda2045324d05c791a4fc312ecceb62954e7740482f8df8928560d63cf273dea595023640179f112de755c79717757ef76962175378d6d87360ddb3f3e5f70f"
)

# tx hash: 9f8741194576a090bc71a3f43a03855950f94278fa121e99203e45967e19a7d0
SIGNATURE_TESTNET_XEM_AS_MOSAIC = unhexlify(
    "1bca7b1b9ffb16d2c2adffa665be072bd2d7a0eafe4a9911dc473500c272905edf3d626274deb52aa490137a276d1fca67ee487079ebf9c09f9faa414f8e7c02"
)


@pytest.mark.skip_t2
class TestMsgNEMSigntx(TrezorTest):

    def test_nem_signtx_simple(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                # Confirm transfer and network fee
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                # Unencrypted message
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                # Confirm recipient
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.NEMSignedTx(signature=SIGNATURE_TESTNET_SIMPLE),
            ])

            self.client.nem_sign_tx(self.client.expand_path("m/44'/1'/0'/0'/0'"), {
                "timeStamp": 74649215,
                "amount": 2000000,
                "fee": 2000000,
                "recipient": "TALICE2GMA34CXHD7XLJQ536NM5UNKQHTORNNT2J",
                "type": 257,
                "deadline": 74735615,
                "message": {
                    "payload": hexlify(b"test_nem_transaction_transfer"),
                    "type": 1,
                },
                "version": (0x98 << 24),
            })

    def test_nem_signtx_xem_as_mosaic(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                # Confirm transfer and network fee
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                # Confirm recipient
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.NEMSignedTx(signature=SIGNATURE_TESTNET_XEM_AS_MOSAIC),
            ])

            self.client.nem_sign_tx(self.client.expand_path("m/44'/1'/0'/0'/0'"), {
                "timeStamp": 76809215,
                "amount": 1000000,
                "fee": 1000000,
                "recipient": "TALICE2GMA34CXHD7XLJQ536NM5UNKQHTORNNT2J",
                "type": 257,
                "deadline": 76895615,
                "version": (0x98 << 24),
                "message": {
                },
                "mosaics": [
                    {
                        "mosaicId": {
                            "namespaceId": "nem",
                            "name": "xem",
                        },
                        "quantity": 1000000,
                    },
                ],
            })



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from .common import *

from hashlib import sha256
from trezorlib import ed25519raw, ed25519cosi


@pytest.mark.skip_t2
class TestCosi(TrezorTest):

    def test_cosi_commit(self):
        self.setup_mnemonic_pin_passphrase()

        digest = sha256(b'this is a message').digest()

        c0 = self.client.cosi_commit(self.client.expand_path("10018'/0'"), digest)
        c1 = self.client.cosi_commit(self.client.expand_path("10018'/1'"), digest)
        c2 = self.client.cosi_commit(self.client.expand_path("10018'/2'"), digest)

        assert c0.pubkey != c1.pubkey
        assert c0.pubkey != c2.pubkey
        assert c1.pubkey != c2.pubkey

        assert c0.commitment != c1.commitment
        assert c0.commitment != c2.commitment
        assert c1.commitment != c2.commitment

        digestb = sha256(b'this is a different message').digest()

        c0b = self.client.cosi_commit(self.client.expand_path("10018'/0'"), digestb)
        c1b = self.client.cosi_commit(self.client.expand_path("10018'/1'"), digestb)
        c2b = self.client.cosi_commit(self.client.expand_path("10018'/2'"), digestb)

        assert c0.pubkey == c0b.pubkey
        assert c1.pubkey == c1b.pubkey
        assert c2.pubkey == c2b.pubkey

        assert c0.commitment != c0b.commitment
        assert c1.commitment != c1b.commitment
        assert c2.commitment != c2b.commitment

    def test_cosi_sign(self):
        self.setup_mnemonic_pin_passphrase()

        digest = sha256(b'this is a message').digest()

        c0 = self.client.cosi_commit(self.client.expand_path("10018'/0'"), digest)
        c1 = self.client.cosi_commit(self.client.expand_path("10018'/1'"), digest)
        c2 = self.client.cosi_commit(self.client.expand_path("10018'/2'"), digest)

        global_pk = ed25519cosi.combine_keys([c0.pubkey, c1.pubkey, c2.pubkey])
        global_R = ed25519cosi.combine_keys([c0.commitment, c1.commitment, c2.commitment])

        sig0 = self.client.cosi_sign(self.client.expand_path("10018'/0'"), digest, global_R, global_pk)
        sig1 = self.client.cosi_sign(self.client.expand_path("10018'/1'"), digest, global_R, global_pk)
        sig2 = self.client.cosi_sign(self.client.expand_path("10018'/2'"), digest, global_R, global_pk)

        sig = ed25519cosi.combine_sig(global_R, [sig0.signature, sig1.signature, sig2.signature])

        ed25519raw.checkvalid(sig, digest, global_pk)



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class NEMMosaicDefinition(p.MessageType):
    FIELDS = {
        1: ('name', p.UnicodeType, 0),
        2: ('ticker', p.UnicodeType, 0),
        3: ('namespace', p.UnicodeType, 0),
        4: ('mosaic', p.UnicodeType, 0),
        5: ('divisibility', p.UVarintType, 0),
        6: ('levy', p.UVarintType, 0),
        7: ('fee', p.UVarintType, 0),
        8: ('levy_address', p.UnicodeType, 0),
        9: ('levy_namespace', p.UnicodeType, 0),
        10: ('levy_mosaic', p.UnicodeType, 0),
        11: ('supply', p.UVarintType, 0),
        12: ('mutable_supply', p.BoolType, 0),
        13: ('transferable', p.BoolType, 0),
        14: ('description', p.UnicodeType, 0),
        15: ('networks', p.UVarintType, p.FLAG_REPEATED),
    }



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class Ping(p.MessageType):
    FIELDS = {
        1: ('message', p.UnicodeType, 0),
        2: ('button_protection', p.BoolType, 0),
        3: ('pin_protection', p.BoolType, 0),
        4: ('passphrase_protection', p.BoolType, 0),
    }
    MESSAGE_WIRE_TYPE = 1



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import struct
from io import BytesIO
from . import messages as proto
from . import mapping
from . import protobuf

REPLEN = 64


class ProtocolV2(object):

    def __init__(self):
        self.session = None

    def session_begin(self, transport):
        chunk = struct.pack('>B', 0x03)
        chunk = chunk.ljust(REPLEN, b'\x00')
        transport.write_chunk(chunk)
        resp = transport.read_chunk()
        self.session = self.parse_session_open(resp)

    def session_end(self, transport):
        if not self.session:
            return
        chunk = struct.pack('>BL', 0x04, self.session)
        chunk = chunk.ljust(REPLEN, b'\x00')
        transport.write_chunk(chunk)
        resp = transport.read_chunk()
        (magic, ) = struct.unpack('>B', resp[:1])
        if magic != 0x04:
            raise RuntimeError('Expected session close')
        self.session = None

    def write(self, transport, msg):
        if not self.session:
            raise RuntimeError('Missing session for v2 protocol')

        # Serialize whole message
        data = BytesIO()
        protobuf.dump_message(data, msg)
        data = data.getvalue()
        dataheader = struct.pack('>LL', mapping.get_type(msg), len(data))
        data = dataheader + data
        seq = -1

        # Write it out
        while data:
            if seq < 0:
                repheader = struct.pack('>BL', 0x01, self.session)
            else:
                repheader = struct.pack('>BLL', 0x02, self.session, seq)
            datalen = REPLEN - len(repheader)
            chunk = repheader + data[:datalen]
            chunk = chunk.ljust(REPLEN, b'\x00')
            transport.write_chunk(chunk)
            data = data[datalen:]
            seq += 1

    def read(self, transport):
        if not self.session:
            raise RuntimeError('Missing session for v2 protocol')

        # Read header with first part of message data
        chunk = transport.read_chunk()
        msg_type, datalen, data = self.parse_first(chunk)

        # Read the rest of the message
        while len(data) < datalen:
            chunk = transport.read_chunk()
            next_data = self.parse_next(chunk)
            data.extend(next_data)

        # Strip padding
        data = BytesIO(data[:datalen])

        # Parse to protobuf
        msg = protobuf.load_message(data, mapping.get_class(msg_type))
        return msg

    def parse_first(self, chunk):
        try:
            headerlen = struct.calcsize('>BLLL')
            (magic, session, msg_type, datalen) = struct.unpack('>BLLL', chunk[:headerlen])
        except:
            raise RuntimeError('Cannot parse header')
        if magic != 0x01:
            raise RuntimeError('Unexpected magic character')
        if session != self.session:
            raise RuntimeError('Session id mismatch')
        return msg_type, datalen, chunk[headerlen:]

    def parse_next(self, chunk):
        try:
            headerlen = struct.calcsize('>BLL')
            (magic, session, sequence) = struct.unpack('>BLL', chunk[:headerlen])
        except:
            raise RuntimeError('Cannot parse header')
        if magic != 0x02:
            raise RuntimeError('Unexpected magic characters')
        if session != self.session:
            raise RuntimeError('Session id mismatch')
        return chunk[headerlen:]

    def parse_session_open(self, chunk):
        try:
            headerlen = struct.calcsize('>BL')
            (magic, session) = struct.unpack('>BL', chunk[:headerlen])
        except:
            raise RuntimeError('Cannot parse header')
        if magic != 0x03:
            raise RuntimeError('Unexpected magic character')
        return session



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class HDNodeType(p.MessageType):
    FIELDS = {
        1: ('depth', p.UVarintType, 0),  # required
        2: ('fingerprint', p.UVarintType, 0),  # required
        3: ('child_num', p.UVarintType, 0),  # required
        4: ('chain_code', p.BytesType, 0),  # required
        5: ('private_key', p.BytesType, 0),
        6: ('public_key', p.BytesType, 0),
    }



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
# Copyright (C) 2016      Jochen Hoenicke <hoenicke@gmail.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import requests
import binascii
from io import BytesIO
import struct

from . import mapping
from . import messages
from . import protobuf
from .transport import Transport, TransportException

TREZORD_HOST = 'http://127.0.0.1:21325'


def get_error(resp):
    return ' (error=%d str=%s)' % (resp.status_code, resp.json()['error'])


class BridgeTransport(Transport):
    '''
    BridgeTransport implements transport through TREZOR Bridge (aka trezord).
    '''

    PATH_PREFIX = 'bridge'
    HEADERS = {'Origin': 'https://python.trezor.io'}

    def __init__(self, device):
        super(BridgeTransport, self).__init__()

        self.device = device
        self.conn = requests.Session()
        self.session = None
        self.response = None

    def __str__(self):
        return self.get_path()

    def get_path(self):
        return '%s:%s' % (self.PATH_PREFIX, self.device['path'])

    @classmethod
    def enumerate(cls):
        try:
            r = requests.post(TREZORD_HOST + '/enumerate', headers=cls.HEADERS)
            if r.status_code != 200:
                raise TransportException('trezord: Could not enumerate devices' + get_error(r))
            return [BridgeTransport(dev) for dev in r.json()]
        except:
            return []

    @classmethod
    def find_by_path(cls, path):
        if isinstance(path, bytes):
            path = path.decode()
        path = path.replace('%s:' % cls.PATH_PREFIX, '')

        for transport in BridgeTransport.enumerate():
            if path is None or transport.device['path'] == path:
                return transport
        raise TransportException('Bridge device not found')

    def open(self):
        r = self.conn.post(TREZORD_HOST + '/acquire/%s/null' % self.device['path'], headers=self.HEADERS)
        if r.status_code != 200:
            raise TransportException('trezord: Could not acquire session' + get_error(r))
        self.session = r.json()['session']

    def close(self):
        if not self.session:
            return
        r = self.conn.post(TREZORD_HOST + '/release/%s' % self.session, headers=self.HEADERS)
        if r.status_code != 200:
            raise TransportException('trezord: Could not release session' + get_error(r))
        self.session = None

    def write(self, msg):
        data = BytesIO()
        protobuf.dump_message(data, msg)
        ser = data.getvalue()
        header = struct.pack(">HL", mapping.get_type(msg), len(ser))
        data = binascii.hexlify(header + ser).decode()
        r = self.conn.post(
            TREZORD_HOST + '/call/%s' % self.session, data=data, headers=self.HEADERS)
        if r.status_code != 200:
            raise TransportException('trezord: Could not write message' + get_error(r))
        self.response = r.text

    def read(self):
        if self.response is None:
            raise TransportException('No response stored')
        data = binascii.unhexlify(self.response)
        headerlen = struct.calcsize('>HL')
        (msg_type, datalen) = struct.unpack('>HL', data[:headerlen])
        data = BytesIO(data[headerlen:headerlen + datalen])
        msg = protobuf.load_message(data, mapping.get_class(msg_type))
        self.response = None
        return msg



#!/usr/bin/env python3
import os
import subprocess

from setuptools import setup, Command
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
from setuptools.command.install_scripts import install_scripts

install_requires = [
    'setuptools>=19.0',
    'ecdsa>=0.9',
    'mnemonic>=0.17',
    'requests>=2.4.0',
    'click>=6.2',
    'pyblake2>=0.9.3',
]

import sys
if '--disable-hidapi' in sys.argv:
    sys.argv.remove('--disable-hidapi')
else:
    install_requires.append('hidapi>=0.7.99.post20')

if '--disable-libusb' in sys.argv:
    sys.argv.remove('--disable-libusb')
else:
    install_requires.append('libusb1>=1.6.4')

from trezorlib import __version__ as VERSION


class RenameLib(Command):
    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # Rename the package name into "safetlib", and the commandline tool into "saftctl"
        subprocess.call(["A=trezorlib; B=safetlib; for i in $(grep -R $A -l --exclude-dir=.git ./build); do sed -i -e \"s/$A/$B/g\" $i; done"], shell=True)
        subprocess.call(["A=trezorctl; B=safetctl; for i in $(grep -R $A -l --exclude-dir=.git ./build); do sed -i -e \"s/$A/$B/g\" $i; done"], shell=True)
        subprocess.call(["rm -rf build/lib/safetlib"], shell=True)
        os.rename("build/lib/trezorlib", "build/lib/safetlib")


class RenameScript(Command):
    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # Rename the package name into "safetlib", and the commandline tool into "saftctl"
        subprocess.call(["A=trezorlib; B=safetlib; for i in $(grep -R $A -l --exclude-dir=.git ./build); do sed -i -e \"s/$A/$B/g\" $i; done"], shell=True)
        subprocess.call(["A=trezorctl; B=safetctl; for i in $(grep -R $A -l --exclude-dir=.git ./build); do sed -i -e \"s/$A/$B/g\" $i; done"], shell=True)


def _patch_rename_lib(cls):
    """Patch a setuptools command to depend on `rename_lib`"""
    orig_run = cls.run

    def new_run(self):
        orig_run(self)
        self.run_command('rename_lib')

    cls.run = new_run


def _patch_rename_script(cls):
    """Patch a setuptools command to depend on `rename_script`"""
    orig_run = cls.run

    def new_run(self):
        orig_run(self)
        self.run_command('rename_script')

    cls.run = new_run


_patch_rename_lib(build_py)
_patch_rename_lib(develop)
_patch_rename_script(install_scripts)


setup(
    name='safet',
    version=VERSION,
    author='Trezor - Archos',
    author_email='software@archos.com',
    description='Python library for communicating with the Archos Safe-T mini (Trezor compatible) Hardware Wallets',
    url='https://github.com/archos-safe-t/python-safet',
    packages=[
        'trezorlib',
        'trezorlib.messages',
        'trezorlib.qt',
        'trezorlib.tests.device_tests',
        'trezorlib.tests.unit_tests',
    ],
    scripts=['safetctl'],
    install_requires=install_requires,
    extras_require={
        'ethereum': [
            'rlp>=0.6.0',
        ],
    },
    python_requires='>=3.3',
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3 :: Only',
    ],
    cmdclass={
        'rename_lib': RenameLib,
        'rename_script': RenameScript,
    },
)



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class ResetDevice(p.MessageType):
    FIELDS = {
        1: ('display_random', p.BoolType, 0),
        2: ('strength', p.UVarintType, 0),  # default=256
        3: ('passphrase_protection', p.BoolType, 0),
        4: ('pin_protection', p.BoolType, 0),
        5: ('language', p.UnicodeType, 0),  # default='english'
        6: ('label', p.UnicodeType, 0),
        7: ('u2f_counter', p.UVarintType, 0),
        8: ('skip_backup', p.BoolType, 0),
    }
    MESSAGE_WIRE_TYPE = 14



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .HDNodePathType import HDNodePathType


class MultisigRedeemScriptType(p.MessageType):
    FIELDS = {
        1: ('pubkeys', HDNodePathType, p.FLAG_REPEATED),
        2: ('signatures', p.BytesType, p.FLAG_REPEATED),
        3: ('m', p.UVarintType, 0),
    }



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .MultisigRedeemScriptType import MultisigRedeemScriptType


class TxOutputType(p.MessageType):
    FIELDS = {
        1: ('address', p.UnicodeType, 0),
        2: ('address_n', p.UVarintType, p.FLAG_REPEATED),
        3: ('amount', p.UVarintType, 0),  # required
        4: ('script_type', p.UVarintType, 0),  # required
        5: ('multisig', MultisigRedeemScriptType, 0),
        6: ('op_return_data', p.BytesType, 0),
        7: ('decred_script_version', p.UVarintType, 0),
    }



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .HDNodeType import HDNodeType


class DebugLinkState(p.MessageType):
    FIELDS = {
        1: ('layout', p.BytesType, 0),
        2: ('pin', p.UnicodeType, 0),
        3: ('matrix', p.UnicodeType, 0),
        4: ('mnemonic', p.UnicodeType, 0),
        5: ('node', HDNodeType, 0),
        6: ('passphrase_protection', p.BoolType, 0),
        7: ('reset_word', p.UnicodeType, 0),
        8: ('reset_entropy', p.BytesType, 0),
        9: ('recovery_fake_word', p.UnicodeType, 0),
        10: ('recovery_word_pos', p.UVarintType, 0),
    }
    MESSAGE_WIRE_TYPE = 102



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .NEMCosignatoryModification import NEMCosignatoryModification


class NEMAggregateModification(p.MessageType):
    FIELDS = {
        1: ('modifications', NEMCosignatoryModification, p.FLAG_REPEATED),
        2: ('relative_change', p.Sint32Type, 0),
    }



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .TxInputType import TxInputType
from .TxOutputBinType import TxOutputBinType
from .TxOutputType import TxOutputType


class TransactionType(p.MessageType):
    FIELDS = {
        1: ('version', p.UVarintType, 0),
        2: ('inputs', TxInputType, p.FLAG_REPEATED),
        3: ('bin_outputs', TxOutputBinType, p.FLAG_REPEATED),
        4: ('lock_time', p.UVarintType, 0),
        5: ('outputs', TxOutputType, p.FLAG_REPEATED),
        6: ('inputs_cnt', p.UVarintType, 0),
        7: ('outputs_cnt', p.UVarintType, 0),
        8: ('extra_data', p.BytesType, 0),
        9: ('extra_data_len', p.UVarintType, 0),
        10: ('decred_expiry', p.UVarintType, 0),
        11: ('preblock_hash', p.BytesType, 0),
    }



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .TxRequestDetailsType import TxRequestDetailsType
from .TxRequestSerializedType import TxRequestSerializedType


class TxRequest(p.MessageType):
    FIELDS = {
        1: ('request_type', p.UVarintType, 0),
        2: ('details', TxRequestDetailsType, 0),
        3: ('serialized', TxRequestSerializedType, 0),
    }
    MESSAGE_WIRE_TYPE = 21



# This file is part of the TREZOR project.
#
# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from binascii import hexlify, unhexlify
import pytest
import os

from trezorlib.client import TrezorClient, TrezorClientDebugLink
from trezorlib import tx_api

tests_dir = os.path.dirname(os.path.abspath(__file__))
tx_api.cache_dir = os.path.join(tests_dir, '../txcache')


try:
    from trezorlib.transport_hid import HidTransport
    HID_ENABLED = True
except ImportError as e:
    print('HID transport disabled:', e)
    HID_ENABLED = False

try:
    from trezorlib.transport_webusb import WebUsbTransport
    WEBUSB_ENABLED = True
except ImportError as e:
    print('WebUSB transport disabled:', e)
    WEBUSB_ENABLED = False

try:
    from trezorlib.transport_pipe import PipeTransport
    PIPE_ENABLED = True
except ImportError as e:
    print('PIPE transport disabled:', e)
    PIPE_ENABLED = False

try:
    from trezorlib.transport_udp import UdpTransport
    UDP_ENABLED = True
except ImportError as e:
    print('UDP transport disabled:', e)
    UDP_ENABLED = False


def pipe_exists(path):
    import os
    import stat
    try:
        return stat.S_ISFIFO(os.stat(path).st_mode)
    except:
        return False


def get_transport():
    if HID_ENABLED and HidTransport.enumerate():
        devices = HidTransport.enumerate()
        wirelink = devices[0]
        debuglink = devices[0].find_debug()

    elif WEBUSB_ENABLED and WebUsbTransport.enumerate():
        devices = WebUsbTransport.enumerate()
        wirelink = devices[0]
        debuglink = devices[0].find_debug()

    elif PIPE_ENABLED and pipe_exists('/tmp/pipe.trezor.to'):
        wirelink = PipeTransport('/tmp/pipe.trezor', False)
        debuglink = PipeTransport('/tmp/pipe.trezor_debug', False)

    elif UDP_ENABLED:
        wirelink = UdpTransport('127.0.0.1:21324')
        debuglink = UdpTransport('127.0.0.1:21325')

    return wirelink, debuglink


if HID_ENABLED and HidTransport.enumerate():
    print('Using TREZOR')
elif WEBUSB_ENABLED and WebUsbTransport.enumerate():
    print('Using TREZOR via WebUSB')
elif PIPE_ENABLED and pipe_exists('/tmp/pipe.trezor.to'):
    print('Using Emulator (v1=pipe)')
elif UDP_ENABLED:
    print('Using Emulator (v2=udp)')


class TrezorTest(object):

    def setup_method(self, method):
        wirelink, debuglink = get_transport()
        self.client = TrezorClientDebugLink(wirelink)
        self.client.set_debuglink(debuglink)
        self.client.set_tx_api(tx_api.TxApiBitcoin)
        # self.client.set_buttonwait(3)

        #                     1      2     3    4      5      6      7     8      9    10    11    12
        self.mnemonic12 = 'alcohol woman abuse must during monitor noble actual mixed trade anger aisle'
        self.mnemonic18 = 'owner little vague addict embark decide pink prosper true fork panda embody mixture exchange choose canoe electric jewel'
        self.mnemonic24 = 'dignity pass list indicate nasty swamp pool script soccer toe leaf photo multiply desk host tomato cradle drill spread actor shine dismiss champion exotic'
        self.mnemonic_all = ' '.join(['all'] * 12)

        self.pin4 = '1234'
        self.pin6 = '789456'
        self.pin8 = '45678978'

        self.client.wipe_device()
        self.client.transport.session_begin()

    def teardown_method(self, method):
        self.client.transport.session_end()
        self.client.close()

    def setup_mnemonic_allallall(self):
        self.client.load_device_by_mnemonic(mnemonic=self.mnemonic_all, pin='', passphrase_protection=False, label='test', language='english')

    def setup_mnemonic_nopin_nopassphrase(self):
        self.client.load_device_by_mnemonic(mnemonic=self.mnemonic12, pin='', passphrase_protection=False, label='test', language='english')

    def setup_mnemonic_nopin_passphrase(self):
        self.client.load_device_by_mnemonic(mnemonic=self.mnemonic12, pin='', passphrase_protection=True, label='test', language='english')

    def setup_mnemonic_pin_nopassphrase(self):
        self.client.load_device_by_mnemonic(mnemonic=self.mnemonic12, pin=self.pin4, passphrase_protection=False, label='test', language='english')

    def setup_mnemonic_pin_passphrase(self):
        self.client.load_device_by_mnemonic(mnemonic=self.mnemonic12, pin=self.pin4, passphrase_protection=True, label='test', language='english')


def generate_entropy(strength, internal_entropy, external_entropy):
    '''
    strength - length of produced seed. One of 128, 192, 256
    random - binary stream of random data from external HRNG
    '''
    import hashlib

    if strength not in (128, 192, 256):
        raise ValueError("Invalid strength")

    if not internal_entropy:
        raise ValueError("Internal entropy is not provided")

    if len(internal_entropy) < 32:
        raise ValueError("Internal entropy too short")

    if not external_entropy:
        raise ValueError("External entropy is not provided")

    if len(external_entropy) < 32:
        raise ValueError("External entropy too short")

    entropy = hashlib.sha256(internal_entropy + external_entropy).digest()
    entropy_stripped = entropy[:strength // 8]

    if len(entropy_stripped) * 8 != strength:
        raise ValueError("Entropy length mismatch")

    return entropy_stripped



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p
from .IdentityType import IdentityType


class SignIdentity(p.MessageType):
    FIELDS = {
        1: ('identity', IdentityType, 0),
        2: ('challenge_hidden', p.BytesType, 0),
        3: ('challenge_visual', p.UnicodeType, 0),
        4: ('ecdsa_curve_name', p.UnicodeType, 0),
    }
    MESSAGE_WIRE_TYPE = 53



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class GetPublicKey(p.MessageType):
    FIELDS = {
        1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
        2: ('ecdsa_curve_name', p.UnicodeType, 0),
        3: ('show_display', p.BoolType, 0),
        4: ('coin_name', p.UnicodeType, 0),  # default='Bitcoin'
    }
    MESSAGE_WIRE_TYPE = 11



# Automatically generated by pb2py
from __future__ import absolute_import
from .. import protobuf as p


class IdentityType(p.MessageType):
    FIELDS = {
        1: ('proto', p.UnicodeType, 0),
        2: ('user', p.UnicodeType, 0),
        3: ('host', p.UnicodeType, 0),
        4: ('port', p.UnicodeType, 0),
        5: ('path', p.UnicodeType, 0),
        6: ('index', p.UVarintType, 0),  # default=0
    }
